@@ -114,13 +114,36 @@ struct QQmlMetaTypeData
     QSet<QString> protectedNamespaces;
 
     QString typeRegistrationNamespace;
-    QStringList typeRegistrationFailures;
 
     QHash<int, int> qmlLists;
 
     QHash<const QMetaObject *, QQmlPropertyCache *> propertyCaches;
-    QQmlPropertyCache *propertyCache(const QMetaObject *metaObject);
+    QQmlPropertyCache *propertyCache(const QMetaObject *metaObject, int minorVersion);
     QQmlPropertyCache *propertyCache(const QQmlType &type, int minorVersion);
+
+    void startRecordingTypeRegFailures(QStringList *storage)
+    { typeRegistrationFailures = storage; }
+    void stopRecordingTypeRegFailures()
+    { startRecordingTypeRegFailures(nullptr); }
+    void recordTypeRegFailure(const QString &message)
+    {
+        if (typeRegistrationFailures)
+            typeRegistrationFailures->append(message);
+        else
+            qWarning("%s", message.toUtf8().constData());
+    }
+
+private:
+    QStringList *typeRegistrationFailures = nullptr;
+};
+
+struct EnumInfo {
+    QStringList path;
+    QString metaObjectName;
+    QString enumName;
+    QString enumKey;
+    QString metaEnumScope;
+    bool scoped;
 };
 
 class QQmlTypeModulePrivate
@@ -152,6 +175,16 @@ static uint qHash(const QQmlMetaTypeData::VersionedUri &v)
     return v.uri.hash() ^ qHash(v.majorVersion);
 }
 
+QQmlMetaTypeRegistrationFailureRecorder::QQmlMetaTypeRegistrationFailureRecorder()
+{
+    metaTypeData()->startRecordingTypeRegFailures(&_failures);
+}
+
+QQmlMetaTypeRegistrationFailureRecorder::~QQmlMetaTypeRegistrationFailureRecorder()
+{
+    metaTypeData()->stopRecordingTypeRegFailures();
+}
+
 QQmlMetaTypeData::QQmlMetaTypeData()
 {
 }
@@ -191,9 +224,9 @@ public:
         QQmlCustomParser *customParser;
         QQmlAttachedPropertiesFunc attachedPropertiesFunc;
         const QMetaObject *attachedPropertiesType;
-        int attachedPropertiesId;
         int propertyValueSourceCast;
         int propertyValueInterceptorCast;
+        bool registerEnumClassesUnscoped;
     };
 
     struct QQmlSingletonTypeData
@@ -227,15 +260,14 @@ public:
 
     int index;
     mutable volatile bool isSetup:1;
-    mutable volatile bool isEnumSetup:1;
+    mutable volatile bool isEnumFromCacheSetup:1;
+    mutable volatile bool isEnumFromBaseSetup:1;
     mutable bool haveSuperType:1;
     mutable QList<QQmlProxyMetaObject::ProxyData> metaObjects;
     mutable QStringHash<int> enums;
     mutable QStringHash<int> scopedEnumIndex; // maps from enum name to index in scopedEnums
     mutable QList<QStringHash<int>*> scopedEnums;
 
-    static QHash<const QMetaObject *, int> attachedPropertyIds;
-
     struct PropertyCacheByMinorVersion
     {
         PropertyCacheByMinorVersion() : cache(nullptr), minorVersion(-1) {}
@@ -246,6 +278,9 @@ public:
     QVector<PropertyCacheByMinorVersion> propertyCaches;
     QQmlPropertyCache *propertyCacheForMinorVersion(int minorVersion) const;
     void setPropertyCacheForMinorVersion(int minorVersion, QQmlPropertyCache *cache);
+private:
+    void createListOfPossibleConflictingItems(const QMetaObject *metaObject, QList<EnumInfo> &enumInfoList, QStringList path) const;
+    void createEnumConflictReport(const QMetaObject *metaObject, const QString &conflictingKey) const;
 };
 
 void QQmlMetaTypeData::registerType(QQmlTypePrivate *priv)
@@ -264,7 +299,14 @@ void QQmlMetaTypeData::registerType(QQmlTypePrivate *priv)
 void QQmlType::SingletonInstanceInfo::init(QQmlEngine *e)
 {
     if (scriptCallback && scriptApi(e).isUndefined()) {
-        setScriptApi(e, scriptCallback(e, e));
+        QJSValue value = scriptCallback(e, e);
+        if (value.isQObject()) {
+            QObject *o = value.toQObject();
+            // even though the object is defined in C++, qmlContext(obj) and qmlEngine(obj)
+            // should behave identically to QML singleton types.
+            e->setContextForObject(o, new QQmlContext(e->rootContext(), e));
+        }
+        setScriptApi(e, value);
     } else if (qobjectCallback && !qobjectApi(e)) {
         QObject *o = qobjectCallback(e, e);
         setQObjectApi(e, o);
@@ -273,6 +315,9 @@ void QQmlType::SingletonInstanceInfo::init(QQmlEngine *e)
         }
         // if this object can use a property cache, create it now
         QQmlData::ensurePropertyCache(e, o);
+        // even though the object is defined in C++, qmlContext(obj) and qmlEngine(obj)
+        // should behave identically to QML singleton types.
+        e->setContextForObject(o, new QQmlContext(e->rootContext(), e));
     } else if (!url.isEmpty() && !qobjectApi(e)) {
         QQmlComponent component(e, url, QQmlComponent::PreferSynchronous);
         QObject *o = component.beginCreate(e->rootContext());
@@ -315,12 +360,11 @@ QJSValue QQmlType::SingletonInstanceInfo::scriptApi(QQmlEngine *e) const
     return scriptApis.value(e);
 }
 
-QHash<const QMetaObject *, int> QQmlTypePrivate::attachedPropertyIds;
-
 QQmlTypePrivate::QQmlTypePrivate(QQmlType::RegistrationType type)
 : refCount(1), regType(type), iid(nullptr), typeId(0), listId(0), revision(0),
     containsRevisionedAttributes(false), baseMetaObject(nullptr),
-    index(-1), isSetup(false), isEnumSetup(false), haveSuperType(false)
+    index(-1), isSetup(false), isEnumFromCacheSetup(false), isEnumFromBaseSetup(false),
+    haveSuperType(false)
 {
     switch (type) {
     case QQmlType::CppType:
@@ -335,6 +379,7 @@ QQmlTypePrivate::QQmlTypePrivate(QQmlType::RegistrationType type)
         extraData.cd->attachedPropertiesType = nullptr;
         extraData.cd->propertyValueSourceCast = -1;
         extraData.cd->propertyValueInterceptorCast = -1;
+        extraData.cd->registerEnumClassesUnscoped = true;
         break;
     case QQmlType::SingletonType:
     case QQmlType::CompositeSingletonType:
@@ -448,22 +493,22 @@ QQmlType::QQmlType(QQmlMetaTypeData *data, const QString &elementName, const QQm
     d->baseMetaObject = type.metaObject;
     d->extraData.cd->attachedPropertiesFunc = type.attachedPropertiesFunction;
     d->extraData.cd->attachedPropertiesType = type.attachedPropertiesMetaObject;
-    if (d->extraData.cd->attachedPropertiesType) {
-        QHash<const QMetaObject *, int>::Iterator iter = d->attachedPropertyIds.find(d->baseMetaObject);
-        if (iter == d->attachedPropertyIds.end())
-            iter = d->attachedPropertyIds.insert(d->baseMetaObject, d->index);
-        d->extraData.cd->attachedPropertiesId = *iter;
-    } else {
-        d->extraData.cd->attachedPropertiesId = -1;
-    }
     d->extraData.cd->parserStatusCast = type.parserStatusCast;
     d->extraData.cd->propertyValueSourceCast = type.valueSourceCast;
     d->extraData.cd->propertyValueInterceptorCast = type.valueInterceptorCast;
     d->extraData.cd->extFunc = type.extensionObjectCreate;
     d->extraData.cd->customParser = type.customParser;
+    d->extraData.cd->registerEnumClassesUnscoped = true;
 
     if (type.extensionMetaObject)
         d->extraData.cd->extMetaObject = type.extensionMetaObject;
+
+    // Check if the user wants only scoped enum classes
+    if (d->baseMetaObject) {
+        auto indexOfClassInfo = d->baseMetaObject->indexOfClassInfo("RegisterEnumClassesUnscoped");
+        if (indexOfClassInfo != -1 && QString::fromUtf8(d->baseMetaObject->classInfo(indexOfClassInfo).value()) == QLatin1String("false"))
+            d->extraData.cd->registerEnumClassesUnscoped = false;
+    }
 }
 
 QQmlType::QQmlType(QQmlMetaTypeData *data, const QString &elementName, const QQmlPrivate::RegisterCompositeType &type)
@@ -576,7 +621,7 @@ QQmlType QQmlType::resolveCompositeBaseType(QQmlEnginePrivate *engine) const
     Q_ASSERT(isComposite());
     if (!engine || !d)
         return QQmlType();
-    QQmlRefPointer<QQmlTypeData> td(engine->typeLoader.getType(sourceUrl()), QQmlRefPointer<QQmlTypeData>::Adopt);
+    QQmlRefPointer<QQmlTypeData> td(engine->typeLoader.getType(sourceUrl()));
     if (td.isNull() || !td->isComplete())
         return QQmlType();
     QV4::CompiledData::CompilationUnit *compilationUnit = td->compilationUnit();
@@ -590,11 +635,11 @@ QQmlPropertyCache *QQmlType::compositePropertyCache(QQmlEnginePrivate *engine) c
     Q_ASSERT(isComposite());
     if (!engine)
         return nullptr;
-    QQmlRefPointer<QQmlTypeData> td(engine->typeLoader.getType(sourceUrl()), QQmlRefPointer<QQmlTypeData>::Adopt);
+    QQmlRefPointer<QQmlTypeData> td(engine->typeLoader.getType(sourceUrl()));
     if (td.isNull() || !td->isComplete())
         return nullptr;
     QV4::CompiledData::CompilationUnit *compilationUnit = td->compilationUnit();
-    return compilationUnit->rootPropertyCache();
+    return compilationUnit->rootPropertyCache().data();
 }
 
 static void clone(QMetaObjectBuilder &builder, const QMetaObject *mo,
@@ -765,42 +810,72 @@ void QQmlTypePrivate::init() const
 
 void QQmlTypePrivate::initEnums(const QQmlPropertyCache *cache) const
 {
-    if (isEnumSetup) return;
+    if ((isEnumFromBaseSetup || !baseMetaObject)
+           && (isEnumFromCacheSetup || !cache)) {
+        return;
+    }
 
     init();
 
     QMutexLocker lock(metaTypeDataLock());
-    if (isEnumSetup) return;
 
-    if (cache)
+    if (!isEnumFromCacheSetup && cache) {
         insertEnumsFromPropertyCache(cache);
-    if (baseMetaObject) // could be singleton type without metaobject
-        insertEnums(baseMetaObject);
+        isEnumFromCacheSetup = true;
+    }
 
-    isEnumSetup = true;
+    if (!isEnumFromBaseSetup && baseMetaObject) { // could be singleton type without metaobject
+        insertEnums(baseMetaObject);
+        isEnumFromBaseSetup = true;
+    }
 }
 
 void QQmlTypePrivate::insertEnums(const QMetaObject *metaObject) const
 {
     // Add any enum values defined by 'related' classes
     if (metaObject->d.relatedMetaObjects) {
-        const QMetaObject * const *related = metaObject->d.relatedMetaObjects;
+        const auto *related = metaObject->d.relatedMetaObjects;
         if (related) {
             while (*related)
                 insertEnums(*related++);
         }
     }
 
+    QSet<QString> localEnums;
+    const QMetaObject *localMetaObject = nullptr;
+
     // Add any enum values defined by this class, overwriting any inherited values
     for (int ii = 0; ii < metaObject->enumeratorCount(); ++ii) {
         QMetaEnum e = metaObject->enumerator(ii);
         const bool isScoped = e.isScoped();
         QStringHash<int> *scoped = isScoped ? new QStringHash<int>() : nullptr;
 
+        // We allow enums in sub-classes to overwrite enums from base-classes, such as
+        // ListView.Center (from enum PositionMode) overwriting Item.Center (from enum TransformOrigin).
+        // This is acceptable because the _use_ of the enum from the QML side requires qualification
+        // anyway, i.e. ListView.Center vs. Item.Center.
+        // However if a class defines two enums with the same value, then that must produce a warning
+        // because it represents a valid conflict.
+        if (e.enclosingMetaObject() != localMetaObject) {
+            localEnums.clear();
+            localMetaObject = e.enclosingMetaObject();
+        }
+
         for (int jj = 0; jj < e.keyCount(); ++jj) {
             const QString key = QString::fromUtf8(e.key(jj));
             const int value = e.value(jj);
-            enums.insert(key, value);
+             if (!isScoped || (regType == QQmlType::CppType && extraData.cd->registerEnumClassesUnscoped)) {
+                 if (localEnums.contains(key)) {
+                     auto existingEntry = enums.find(key);
+                     if (existingEntry != enums.end() && existingEntry.value() != value) {
+                         qWarning("Previously registered enum will be overwritten due to name clash: %s.%s", metaObject->className(), key.toUtf8().constData());
+                         createEnumConflictReport(metaObject, key);
+                     }
+                 } else {
+                     localEnums.insert(key);
+                 }
+                 enums.insert(key, value);
+            }
             if (isScoped)
                 scoped->insert(key, value);
         }
@@ -812,6 +887,59 @@ void QQmlTypePrivate::insertEnums(const QMetaObject *metaObject) const
     }
 }
 
+void QQmlTypePrivate::createListOfPossibleConflictingItems(const QMetaObject *metaObject, QList<EnumInfo> &enumInfoList, QStringList path) const
+{
+    path.append(QString::fromUtf8(metaObject->className()));
+
+    if (metaObject->d.relatedMetaObjects) {
+        const auto *related = metaObject->d.relatedMetaObjects;
+        if (related) {
+            while (*related)
+                createListOfPossibleConflictingItems(*related++, enumInfoList, path);
+        }
+    }
+
+    for (int ii = 0; ii < metaObject->enumeratorCount(); ++ii) {
+        const auto e = metaObject->enumerator(ii);
+
+        for (int jj = 0; jj < e.keyCount(); ++jj) {
+            const QString key = QString::fromUtf8(e.key(jj));
+
+            EnumInfo enumInfo;
+            enumInfo.metaObjectName = QString::fromUtf8(metaObject->className());
+            enumInfo.enumName = QString::fromUtf8(e.name());
+            enumInfo.enumKey = key;
+            enumInfo.scoped = e.isScoped();
+            enumInfo.path = path;
+            enumInfo.metaEnumScope = QString::fromUtf8(e.scope());
+            enumInfoList.append(enumInfo);
+        }
+    }
+}
+
+void QQmlTypePrivate::createEnumConflictReport(const QMetaObject *metaObject, const QString &conflictingKey) const
+{
+    QList<EnumInfo> enumInfoList;
+
+    if (baseMetaObject) // prefer baseMetaObject if available
+        metaObject = baseMetaObject;
+
+    if (!metaObject) { // If there is no metaObject at all return early
+        qWarning() << "No meta object information available. Skipping conflict analysis.";
+        return;
+    }
+
+    createListOfPossibleConflictingItems(metaObject, enumInfoList, QStringList());
+
+    qWarning().noquote() << QLatin1String("Possible conflicting items:");
+    // find items with conflicting key
+    for (const auto i : enumInfoList) {
+        if (i.enumKey == conflictingKey)
+            qWarning().noquote().nospace() << "    " << i.metaObjectName << "." << i.enumName << "." << i.enumKey << " from scope "
+                                           << i.metaEnumScope << " injected by " << i.path.join(QLatin1String("->"));
+    }
+}
+
 void QQmlTypePrivate::insertEnumsFromPropertyCache(const QQmlPropertyCache *cache) const
 {
     const QMetaObject *cppMetaObject = cache->firstCppMetaObject();
@@ -841,7 +969,7 @@ QQmlPropertyCache *QQmlTypePrivate::propertyCacheForMinorVersion(int minorVersio
 {
     for (int i = 0; i < propertyCaches.count(); ++i)
         if (propertyCaches.at(i).minorVersion == minorVersion)
-            return propertyCaches.at(i).cache;
+            return propertyCaches.at(i).cache.data();
     return nullptr;
 }
 
@@ -965,6 +1093,13 @@ bool QQmlType::isCreatable() const
     return d && d->regType == CppType && d->extraData.cd->newFunc;
 }
 
+QQmlType::ExtensionFunc QQmlType::extensionFunction() const
+{
+    if (!d || d->regType != CppType)
+        return nullptr;
+    return d->extraData.cd->extFunc;
+}
+
 bool QQmlType::isExtendedType() const
 {
     if (!d)
@@ -1072,7 +1207,7 @@ int QQmlType::attachedPropertiesId(QQmlEnginePrivate *engine) const
     if (!d)
         return -1;
     if (d->regType == CppType)
-        return d->extraData.cd->attachedPropertiesId;
+        return d->extraData.cd->attachedPropertiesType ? d->index : -1;
 
     QQmlType base;
     if (d->regType == CompositeType)
@@ -1571,7 +1706,7 @@ bool checkRegistration(QQmlType::RegistrationType typeType, QQmlMetaTypeData *da
     if (!typeName.isEmpty()) {
         if (typeName.at(0).isLower()) {
             QString failure(QCoreApplication::translate("qmlRegisterType", "Invalid QML %1 name \"%2\"; type names must begin with an uppercase letter"));
-            data->typeRegistrationFailures.append(failure.arg(registrationTypeString(typeType)).arg(typeName));
+            data->recordTypeRegFailure(failure.arg(registrationTypeString(typeType)).arg(typeName));
             return false;
         }
 
@@ -1579,7 +1714,7 @@ bool checkRegistration(QQmlType::RegistrationType typeType, QQmlMetaTypeData *da
         for (int ii = 0; ii < typeNameLen; ++ii) {
             if (!(typeName.at(ii).isLetterOrNumber() || typeName.at(ii) == '_')) {
                 QString failure(QCoreApplication::translate("qmlRegisterType", "Invalid QML %1 name \"%2\""));
-                data->typeRegistrationFailures.append(failure.arg(registrationTypeString(typeType)).arg(typeName));
+                data->recordTypeRegFailure(failure.arg(registrationTypeString(typeType)).arg(typeName));
                 return false;
             }
         }
@@ -1593,7 +1728,7 @@ bool checkRegistration(QQmlType::RegistrationType typeType, QQmlMetaTypeData *da
             if (data->protectedNamespaces.contains(nameSpace)) {
                 QString failure(QCoreApplication::translate("qmlRegisterType",
                                                             "Cannot install %1 '%2' into protected namespace '%3'"));
-                data->typeRegistrationFailures.append(failure.arg(registrationTypeString(typeType)).arg(typeName).arg(nameSpace));
+                data->recordTypeRegFailure(failure.arg(registrationTypeString(typeType)).arg(typeName).arg(nameSpace));
                 return false;
             }
         } else if (majorVersion >= 0) {
@@ -1604,7 +1739,7 @@ bool checkRegistration(QQmlType::RegistrationType typeType, QQmlMetaTypeData *da
                 if (QQmlTypeModulePrivate::get(qqtm)->locked){
                     QString failure(QCoreApplication::translate("qmlRegisterType",
                                                                 "Cannot install %1 '%2' into protected module '%3' version '%4'"));
-                    data->typeRegistrationFailures.append(failure.arg(registrationTypeString(typeType)).arg(typeName).arg(nameSpace).arg(majorVersion));
+                    data->recordTypeRegFailure(failure.arg(registrationTypeString(typeType)).arg(typeName).arg(nameSpace).arg(majorVersion));
                     return false;
                 }
             }
@@ -1853,6 +1988,23 @@ void qmlRegisterModule(const char *uri, int versionMajor, int versionMinor)
     p->maxMinorVersion = qMax(p->maxMinorVersion, versionMinor);
 }
 
+//From qqml.h
+int qmlTypeId(const char *uri, int versionMajor, int versionMinor, const char *qmlName)
+{
+    QMutexLocker lock(metaTypeDataLock());
+    QQmlMetaTypeData *data = metaTypeData();
+
+    QQmlTypeModule *module = getTypeModule(QString::fromUtf8(uri), versionMajor, data);
+    if (!module)
+        return -1;
+
+    QQmlType type = module->type(QHashedStringRef(QString::fromUtf8(qmlName)), versionMinor);
+    if (!type.isValid())
+        return -1;
+
+    return type.index();
+}
+
 bool QQmlMetaType::namespaceContainsRegistrations(const QString &uri, int majorVersion)
 {
     const QQmlMetaTypeData *data = metaTypeData();
@@ -1878,14 +2030,6 @@ void QQmlMetaType::setTypeRegistrationNamespace(const QString &uri)
     QQmlMetaTypeData *data = metaTypeData();
 
     data->typeRegistrationNamespace = uri;
-    data->typeRegistrationFailures.clear();
-}
-
-QStringList QQmlMetaType::typeRegistrationFailures()
-{
-    QQmlMetaTypeData *data = metaTypeData();
-
-    return data->typeRegistrationFailures;
 }
 
 QMutex *QQmlMetaType::typeRegistrationLock()
@@ -2023,6 +2167,16 @@ QQmlAttachedPropertiesFunc QQmlMetaType::attachedPropertiesFuncById(QQmlEnginePr
     return data->types.at(id).attachedPropertiesFunction(engine);
 }
 
+QQmlAttachedPropertiesFunc QQmlMetaType::attachedPropertiesFunc(QQmlEnginePrivate *engine,
+                                                                const QMetaObject *mo)
+{
+    QMutexLocker lock(metaTypeDataLock());
+    QQmlMetaTypeData *data = metaTypeData();
+
+    QQmlType type(data->metaObjectToType.value(mo));
+    return type.attachedPropertiesFunction(engine);
+}
+
 QMetaProperty QQmlMetaType::defaultProperty(const QMetaObject *metaObject)
 {
     int idx = metaObject->indexOfClassInfo("DefaultProperty");
@@ -2094,6 +2248,9 @@ QQmlMetaType::TypeCategory QQmlMetaType::typeCategory(int userType)
         return Unknown;
 }
 
+/*!
+    See qmlRegisterInterface() for information about when this will return true.
+*/
 bool QQmlMetaType::isInterface(int userType)
 {
     QMutexLocker lock(metaTypeDataLock());
@@ -2232,19 +2389,25 @@ QQmlType QQmlMetaType::qmlType(const QMetaObject *metaObject, const QHashedStrin
 }
 
 /*!
-    Returns the type (if any) that corresponds to the QVariant::Type \a userType.
-    Returns null if no type is registered.
+    Returns the type (if any) that corresponds to \a typeId.  Depending on \a category, the
+    \a typeId is interpreted either as QVariant::Type or as QML type id returned by one of the
+    qml type registration functions.  Returns null if no type is registered.
 */
-QQmlType QQmlMetaType::qmlType(int userType)
+QQmlType QQmlMetaType::qmlType(int typeId, TypeIdCategory category)
 {
     QMutexLocker lock(metaTypeDataLock());
     QQmlMetaTypeData *data = metaTypeData();
 
-    QQmlTypePrivate *type = data->idToType.value(userType);
-    if (type && type->typeId == userType)
-        return QQmlType(type);
-    else
-        return QQmlType();
+    if (category == TypeIdCategory::MetaType) {
+        QQmlTypePrivate *type = data->idToType.value(typeId);
+        if (type && type->typeId == typeId)
+            return QQmlType(type);
+    } else if (category == TypeIdCategory::QmlType) {
+        QQmlType type = data->types.value(typeId);
+        if (type.isValid())
+            return type;
+    }
+    return QQmlType();
 }
 
 /*!
@@ -2269,7 +2432,7 @@ QQmlType QQmlMetaType::qmlType(const QUrl &unNormalizedUrl, bool includeNonFileI
         return QQmlType();
 }
 
-QQmlPropertyCache *QQmlMetaTypeData::propertyCache(const QMetaObject *metaObject)
+QQmlPropertyCache *QQmlMetaTypeData::propertyCache(const QMetaObject *metaObject, int minorVersion)
 {
     if (QQmlPropertyCache *rv = propertyCaches.value(metaObject))
         return rv;
@@ -2279,17 +2442,17 @@ QQmlPropertyCache *QQmlMetaTypeData::propertyCache(const QMetaObject *metaObject
         propertyCaches.insert(metaObject, rv);
         return rv;
     }
-    QQmlPropertyCache *super = propertyCache(metaObject->superClass());
-    QQmlPropertyCache *rv = super->copyAndAppend(metaObject);
+    QQmlPropertyCache *super = propertyCache(metaObject->superClass(), minorVersion);
+    QQmlPropertyCache *rv = super->copyAndAppend(metaObject, minorVersion);
     propertyCaches.insert(metaObject, rv);
     return rv;
 }
 
-QQmlPropertyCache *QQmlMetaType::propertyCache(const QMetaObject *metaObject)
+QQmlPropertyCache *QQmlMetaType::propertyCache(const QMetaObject *metaObject, int minorVersion)
 {
     QMutexLocker lock(metaTypeDataLock());
     QQmlMetaTypeData *data = metaTypeData();
-    return data->propertyCache(metaObject);
+    return data->propertyCache(metaObject, minorVersion);
 }
 
 QQmlPropertyCache *QQmlMetaTypeData::propertyCache(const QQmlType &type, int minorVersion)
@@ -2322,7 +2485,7 @@ QQmlPropertyCache *QQmlMetaTypeData::propertyCache(const QQmlType &type, int min
         return pc;
     }
 
-    QQmlPropertyCache *raw = propertyCache(type.metaObject());
+    QQmlPropertyCache *raw = propertyCache(type.metaObject(), minorVersion);
 
     bool hasCopied = false;
 
@@ -2430,6 +2593,10 @@ void QQmlMetaType::freeUnusedTypesAndCaches()
     QMutexLocker lock(metaTypeDataLock());
     QQmlMetaTypeData *data = metaTypeData();
 
+    // in case this is being called during program exit, `data` might be destructed already
+    if (!data)
+        return;
+
     {
         bool deletedAtLeastOneType;
         do {
